BeginPackage["ApproxSum`"];

ApproxSum::usage = "ApproxSum[s_,x_,{n_,n0_,Infinity}] approximates the sum"<>
" over s, where s depends on the variable x. Above a cut-off specified "<>
" by the options an expansion of s for n -> Infinity is used";

ApproxSum::exp = "Value `1` of option ExpansionOrder is not an integer number";
ApproxSum::noconv = "Bad apparent convergence of sum. "<>
"Obtained `1` for the sum and `2` for the error estimate"

NApproxSum::usage = "NApproxSum[s_,x_,{n_,n0_,Infinity}] approximates the sum"<>
" over s, where s depends on the variable x. Above a cut-off specified "<>
" by the options the remainder of the summand is approximated numerically. "<>
" In case of nested sums the cut-off refers to the outermost summation index";

NApproxSum::asy = "Bad asymptotic behaviour of sum. Obtained `1` above cut-off and "<>
"`2` below cut-off";

Options[ApproxSum] = {
   ExpansionOrder -> 8,
   CutOff -> (10*Abs[#]&),
   AsymptoticPrecision -> 20
};

Options[NApproxSum] = {
   CutOff -> (10*Abs[#]&),
   AsymptoticPrecision -> MachinePrecision
};

Begin["`Private`"];
Needs["horner`"];


ApproxSum[s_,x_Symbol,{n_Symbol,n0_Integer,Infinity},OptionsPattern[]] := Module[
   {
      badratio = 10^-1,
      sum,
      summand,
      asymptotic,
      c,k,k0=n0,
      exporder = OptionValue[ExpansionOrder],
      cut = OptionValue[CutOff],
      prec = OptionValue[AsymptoticPrecision]
   },
   If[Head[exporder] =!= Integer,
      Message[ApproxSum::exp,exporder];
      Return[Null];
   ];
   asymptotic[var_,c_] = SumAsymptotic[s,{n,exporder}][c] /. x -> var;
   summand[var_] = s /. x -> var /. n -> k;

   sum[var_?NumberQ] := Module[
      {
         c = Ceiling[cut[var]],
         maxRelError = 0.001,
         errorc,
         summandn = summand[var],
         result, error
      },
      errorc = Min[Floor[0.9*c],c-1];
      If[c === Infinity,
         Return[NSum[summandn,{k,k0,c}]];
      ];
      error = Sum[summandn,{k,k0,errorc}];
      result = (
         + error + Sum[summandn,{k,Max[k0,errorc+1],c}] \
         + asymptotic[SetPrecision[var,prec],c+1]
      );
      error += asymptotic[SetPrecision[var,prec],errorc+1];
      error -= result;
      error = Abs /@ {Re[error],Im[error]};
      If[ (error[[1]]/Abs[Re[result]] > maxRelError)
         || (error[[2]]/Abs[Im[result]] > maxRelError),
         Message[ApproxSum::noconv,result,Complex@@error];
      ];
      Return[result];
   ];
   Return[sum[x]];
];

NApproxSum[s_,var_?NumberQ,{n_Symbol,n0_Integer,Infinity},lim___List,OptionsPattern[]] :=
Module[
   {
      badratio = 10^-1,
      c,exact,rest,
      cut = OptionValue[CutOff],
      prec = OptionValue[AsymptoticPrecision]
   },
   c = Ceiling[cut[var]];
   If[c === Infinity,
      Return[NSum[s,{n,n0,c}]];
   ];
   exact = Sum[s, {n,n0,c}, lim];
   rest = NSum[s, {n,c+1,Infinity}, lim, WorkingPrecision -> prec];
   If[ (Abs[Re[rest]] > badratio*Abs[Re[exact]])
      || (Abs[Im[rest]] > badratio*Abs[Im[exact]]),
      Message[NApproxSum::asy,rest,exact]
   ];
   Return[exact+rest];
];


SumAsymptotic::usage = "SumAsymptotic[summand_,{n_,pow_}] returns the sum " <>
"over the summand expanded for large n up to 1/n^pow. "<>
"After f = SumAsymptotic[summand_,{n_,pow_}]; f[n0] is equivalent to "<>
"Sum[ExpandedSummand,{n,n0,Infinity}]";

SumAsymptotic::nosum = "Failed to compute Sum[`1`,{`2`,0,cut}]"

SumAsymptotic[summand_,{n_,pow_Integer}] := Module[
   {summandExp,sum,n0,y,x,ep,tmp},
   summandExp = Collect[
      Normal[Series[
            summand, {n, Infinity, pow}
      ]] /. Log[1/n] -> -Log[n]
      , {n, Log[n]}
      , mask
   ];

   sumCut = Expand[summandExp] \
      /. Log[n]^x_*n^y_ :> (-1)^x*Derivative[x,0][Zeta][-y,n0] \
      /. Log[n]*n^y_ :> -Derivative[1,0][Zeta][-y,n0] \
      /. n^y_ :> Zeta[-y,n0] \
      /. mask[x_] -> x
   ;
   If[!FreeQ[sumCut,n],
      Message[SumAsymptotic::nosum,summandExp /. mask[x_] -> x,n];
      Return[Null];
   ];

   sum[x_] = Collect[
      sumCut /. n0 -> x,
      Cases[bar, Zeta[__] | Derivative[__][Zeta][__], {0, Infinity}],
      Simplify[horner[#]]&
   ];

   Return[sum];
];

End[];
EndPackage[];
